#!/usr/bin/perl
use lib "blib/lib";

#
# Monolithic EXE and PAR distributions generator script: substitutes
# for both a BAT and an UNIX shell script, and for the tar utility as
# well (Archive::Tar is a core module).
#
# Antonio García Domínguez
# $Id$
#

use strict;
use warnings;
use English;
use Config;
use Fatal;

use Archive::Tar;
use File::Temp qw(tempdir);
use File::Copy qw(copy);
use File::Spec::Functions;
use File::stat;
use ACL2::Procesador;

### CONFIGURATION

use constant {
    INCLUDE_DIRS  => ["xslt", "share"],
    INCLUDE_FILTER=> qr/^[^.].*[^~]$/,
    SAMPLE_DIRS   => ["t/testInputs"],
    SAMPLE_FILTER => qr/(?:lisp|lsp|acl2)$/,
    MAIN_SCRIPT   => "script/pprocACL2",
    MODULE_NAME   => "ACL2-Procesador",
    BUILD_ARCHSPECIFIC_PAR => 0,
    TYPE_FILE     => "formats/acl2.format",
    VERSION       => $ACL2::Procesador::VERSION,
};

### MAIN BODY

# List all files of interest for the distributions
my @common_files = (recursive_filtered_list('filter' => INCLUDE_FILTER,
                                            'rv_list_queue' => INCLUDE_DIRS),
                    recursive_filtered_list('filter' => SAMPLE_FILTER,
                                            'rv_list_queue' => SAMPLE_DIRS));
@common_files = map { $_ => testcase_to_sample($_) } @common_files;

# EXE generation
my $path_to_exe = "pprocACL2";
my $exe_command;
if ($OSNAME =~ /^(?:MSWin|cygwin|OS2)/) {
    $path_to_exe .= ".exe";
    $exe_command  = "$path_to_exe \"%s\"";
}
else {
    $exe_command  = "./$path_to_exe \"%s\"";
}
system("pp -r -I blib/lib -M PerlIO " 
       . " -o $path_to_exe "
       . MAIN_SCRIPT);

# PAR generation
my $path_to_pl = "pprocACL2.pl";
my $pl_command = "perl $path_to_pl \"%s\"";
system("pp -I blib/lib -M PerlIO -P"
       . " -o $path_to_pl "
       . MAIN_SCRIPT);

# Variants to be packaged
my @variants = (
    {
        'exec_file'   => $path_to_exe,
        'command'     => $exe_command,
        'name'        => MODULE_NAME
            . "-" . VERSION . "-standalone-$Config{archname}"},
    {
        'exec_file'   => $path_to_pl,
        'command'     => $pl_command,
        'name'        => MODULE_NAME
            . "-" . VERSION . "-par-"
            . (BUILD_ARCHSPECIFIC_PAR ? $Config{archname} : "noarch")}
    );

foreach my $variant (@variants) {
    my $tar       = Archive::Tar->new;

    # Add common and variant-specific files
    my %files = (@common_files,
                 $variant->{exec_file} => $variant->{exec_file});
    while (my ($from, $to) = each(%files)) {
        if ($from eq $variant->{exec_file} && $from !~ /\.par$/) {
            $tar->add_data($to, load_text($from), { mode => 0755 });
        }
        else {
            $tar->add_data($to, load_text($from));
        }
    }

    # Add a modified version of the format descriptor
    my $command = $variant->{command};

    my @lines = load_lines(TYPE_FILE);
    @lines    = map {
        s#(<import_cmd>).*(</import_cmd>)#$1$command$2#; $_
    } @lines;
    $tar->add_data(TYPE_FILE, join("", @lines));

    # Write the tar.gz to disk
    $tar->write($variant->{name} . ".tar.gz", 1);
}

### UTILITY FUNCTIONS

sub recursive_filtered_list
{
    my %args = @_;

    my $filter_pattern = $args{filter};
    my @list_queue     = @{$args{rv_list_queue}};
    my @results        = ();

    while (@list_queue > 0)
    {
        # Relative path from the current directory
        my $path = shift @list_queue;

        if (-d $path) {
            my $olddir;
            opendir($olddir, $path)
                or die "Couldn't open $path: $!";
            my @entries = map { catfile($path, $_) } readdir($olddir);
            closedir($olddir);

            # Add filtered entries to queue (and directories)
            my @filtered = grep {
                ((-f $_ && $_ =~ $filter_pattern) ||
                 (-d $_ && $_ !~ m#(?:^|/)\.{1,2}$#))
            } @entries;
            push @list_queue, @filtered;
        }

        if (-f $path) {
            push @results, $path;
        }
    }

    return @results;
}

sub load_lines
{
    my ($path) = @_;

    local *TYPE_FH;
    open(TYPE_FH, "<", $path)
        or die "Couldn't open " . TYPE_FILE . ": $!";
    my @lines = <TYPE_FH>;
    close(TYPE_FH);

    return @lines;
}

sub load_text
{
    return join "", load_lines(@_);
}

sub testcase_to_sample
{
    my ($test_input_path) = @_;

    my $sample_path = $test_input_path;
    $sample_path =~ s#^t/testInputs/(.*)#samples/$1#;
    return $sample_path;
}
