package ACL2::State;

#
# Global ACL2 state table
# Copyright (C) 2009 Antonio García Domínguez
# $Id$
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use utf8;

use Cwd qw(realpath getcwd);
use File::Temp qw(tempfile);

use ACL2::Localizacion;

use constant {
    # Key to the distributed books directory (see INCLUDE-BOOK in the
    # ACL2 manual)
    BOOK_DIR_NAME_DBD => ":SYSTEM",
    # Name of the environment variable ACL2 uses for the distributed
    # books directory
    DBD_ENV_VAR_NAME => "ACL2_SYSTEM_BOOKS",
};

our $dbd_path;
our $lh = ACL2::Localizacion->get_handle();

=head1 DESCRIPTION

ACL2 has a concept of "state" which contains all the global variables
which control the behaviour of some of the events. Objects of this
class try to emulate part of this behaviour (though error checking
will be quite a bit more lenient).

=head1 FUNCTIONS

All public methods in this class are listed below.

=head2 new

Creates an empty ACL2 state table.

The distributed book directory path will be obtained the first time
any ACL2::State object is created in this project. The only reliable
way across different platforms, ACL2 versions, and user environments
is to ask ACL2 itself. We will need the distributed book directory
_before_ we can produce ACL2's output for the proof script, though.

To solve this problem, we will first try to ask ACL2 itself in a short
one-line session, and if that does not work (ACL2 might not be
installed, and we only have the .out files) fall back on the
ACL2_SYSTEM_BOOKS environment variable. If that isn't set either, we
will signal an error.

=cut

sub new {
    my ($pkg) = @_;

    # We only have to call ACL2 to ask the DBD the first time we
    # create an ACL2::State object in a multifile project
    $dbd_path = _get_dbd_path()
        unless defined($dbd_path);

    return bless {
        "cbd"       => getcwd(),
        "book_dirs" => {BOOK_DIR_NAME_DBD() => $dbd_path},
    }, $pkg;
}

=head2 set_book_dir_by_name($name, $path)

Changes the path of the book directory under the C<name> keyword to
C<path>. C<path> will be normalized to an absolute path.

=cut

sub set_book_dir_by_name {
    my ($this, $name, $path) = @_;
    $this->{book_dirs}->{$name} = realpath $_[1];
}

=head2 get_book_dir_by_name($name)

Returns the path to the book directory under the C<name> keyword, if
it has been set. Otherwise, it will return an undefined value.

=cut

sub get_book_dir_by_name {
    return $_[0]->{book_dirs}->{$_[1]};
}

=head2 set_cbd($path)

Sets the path to the current book directory to C<path>. By default, it
is set to the current working directory at the moment that this State
object was created.

=cut

sub set_cbd {
    my ($this, $path) = @_;
    $this->{cbd} = $path;
}

=head2 get_cbd

Returns the path to the current book directory.

=cut

sub get_cbd {
    return $_[0]->{cbd};
}

=head2 _get_dbd_path

Calls ACL2 to obtain the distributed book directory. If that fails, it
falls back on the environment variable referred to by
C<DBD_ENV_VAR_NAME()>. If that doesn't work either, it signals an
error.

=cut

sub _get_dbd_path {

    # Check the value of the distributed books directory (the first
    # operand is needed in ACL2 2.9, which did not offer the
    # distributed-books-dir global variable).
    my $ask_dbd_script = <<'    SCRIPT';
    (or (cdr (assoc-eq
              :system
              (cdr (assoc-eq
                    :include-book-dir-alist
                    (table-alist (quote acl2-defaults-table)
                                 (w state))))))
        (@ distributed-books-dir))
    SCRIPT

    # First, check for the environment variable
    if (exists $ENV{DBD_ENV_VAR_NAME()}) {
        return $ENV{DBD_ENV_VAR_NAME()};
    }

    # Otherwise, try calling ACL2

    # Dump the script to a temporary file
    my ($fh, $tmpFilename) = tempfile();
    print $fh $ask_dbd_script;
    close $fh;

    # Call ACL2 and obtain the value of the distributed-books-dir variable
    my $dbd_path;
    my $ask_dbd_output = `acl2 < $tmpFilename`;
    if (length($ask_dbd_output) > 0) {
        my @lines          = split "\n", $ask_dbd_output;
        my @prompt_lines   = grep /^ACL2 !>/, @lines;
        if ($prompt_lines[0] =~ /^ACL2 !>"([^"]+)"/) {
            $dbd_path = substr $prompt_lines[0], $-[1], $+[1] - $-[1];
        }
    }

    die $lh->maketext(
        "Could not obtain the path to the distributed books directory "
        . "from ACL2, and the environment variable [_1] is not set",
        DBD_ENV_VAR_NAME())
        unless defined($dbd_path) && length($dbd_path) > 0;

    return $dbd_path;
}

1;

