#!perl -T
# -*- perl -*-

#
# Test examples for the CLTL-based parser
# Antonio García Domínguez
# $Id$
#
use strict;
use warnings;
use Test::More;

use ACL2::Lisp::Parser qw(read_lisp_program);

use constant {
    "REPEAT" => "repeat",
};

##### UTILITIES

# Generate functions for easily creating expected parsing trees
foreach my $elem (qw(Program List)) {
    no strict;
    *{$elem} = sub {
        my $pkg = eval "ACL2::Lisp::$elem";
        if (ref $_[0] eq "ARRAY") {
            # Specify source manually: List([x,y], "source")
            return $pkg->new(@_);
        }
        else {
            # Source will be generated: List(x, y)
            return $pkg->new([@_]);
        }
    };
}
foreach my $elem (qw(Whitespace Symbol
                     String Number
                     Comment
                     Quote Comma Backquote Dot)) {
    no strict;
    *{$elem} = sub {
        my $pkg = eval "ACL2::Lisp::$elem";
        return $pkg->new(@_);
    };
}

##### TEST SPECIFICATION

my @cases = (
    # Empty input
    "", Program(), REPEAT,

    # Whitespace
    " " , Program(Whitespace(" ")),
        , Program([], " "),

    # Symbols and spacing
    " \tx \n"  , Program(Whitespace(" \t"),
                         Symbol("X", "x"),
                         Whitespace(" \n")),
                 Program(Symbol("X", "x")),
    "|hi all|" , Program(Symbol("hi all", "|hi all|")),
                 REPEAT,
    "|hi\nall|", Program(Symbol("hi\nall", "|hi\nall|")),
                 REPEAT,
    '\hI'      , Program(Symbol('hI'  , '\hI')) , REPEAT,
    'h\i'      , Program(Symbol('Hi'  , 'h\i')) , REPEAT,
    'H\\\\i'   , Program(Symbol('H\\I', 'H\\\\i')), REPEAT,
    'H\ i'     , Program(Symbol('H I' , 'H\ i')), REPEAT,
    'acons$'   , Program(Symbol('ACONS$', 'acons$')), REPEAT,

    # Numbers
    "10"    , Program(Number("10")),
              REPEAT,
    "-10."  , Program(Number("-10.")),
              REPEAT,
    "2e2"   , Program(Number("2e2")),
              REPEAT,
    ".3"    , Program(Number(".3")),
              REPEAT,
    ".3e-3" , Program(Number(".3e-3")),
              REPEAT,
    "2.3e+4", Program(Number("2.3e+4")),
              REPEAT,

    # Strings
    '"test"'        , Program(String('test')),
                      REPEAT,
    "\"test\nthis\"", Program(String("test\nthis")),
                      REPEAT,

    # Lists
    '()'     , Program(List())
             , REPEAT,
    '( )'    , Program(List(Whitespace(" ")))
             , Program(List([], "( )")),
    '(x)'    , Program(List(Symbol("X", "x"))),
             , REPEAT,
    '(2 x)'  , Program(List(Number("2"),
                            Whitespace(" "),
                            Symbol("X", "x"))),
             , Program(List([Number("2"),
                             Symbol("X", "x")],
                            "(2 x)")),
    '(2 . x)', Program(List(Number("2"),
                            Whitespace(" "),
                            Symbol('.'),
                            Whitespace(" "),
                            Symbol("X", "x"))),
             , Program(List([Number("2"),
                             Symbol('.'),
                             Symbol("X", "x")],
                            "(2 . x)")),

    # Line comments
    '; comment'        , Program(Comment("; comment"))
                       , Program(),
    "; first\n; second", Program(Comment("; first"),
                                 Whitespace("\n"),
                                 Comment("; second")),
                       , Program(),
    "x ; comment"      , Program(Symbol("X", "x"),
                                 Whitespace(" "),
                                 Comment("; comment")),
                       , Program(Symbol("X", "x")),

    # Block comments
    "#|CommenT |#"   , Program(Comment("#|CommenT |#")),
                     , Program(),
    "#|\n| Test\n|#" , Program(Comment("#|\n| Test\n|#")),
                     , Program(),
    "x #| block |# y", Program(Symbol("X", "x"),
                               Whitespace(" "),
                               Comment("#| block |#"),
                               Whitespace(" "),
                               Symbol("Y", "y")),
                     , Program(Symbol("X", "x"),
                               Symbol("Y", "y")),

    # Quoting
    "'x"     , Program(List([Symbol("QUOTE"),
                             Symbol("X", "x")],
                            "'x")),
             , REPEAT,
    "`z"     , Program(Backquote(Symbol("Z", "z"))),
               REPEAT,
    "`(x ,y)", Program(Backquote(
                           List(Symbol("X", "x"),
                                Whitespace(" "),
                                Comma(Symbol("Y", "y"))))),
               Program(Backquote(
                           List([Symbol("X", "x"),
                                 Comma(Symbol("Y", "y"))],
                                "(x ,y)"))),

);

# Avisamos a Test::More del número de pruebas a realizar
plan tests => scalar @cases;

while(scalar @cases >= 2) {
    my ($input, $expected, $expected_cleaned) = splice @cases, 0, 3;
    $expected_cleaned = $expected if $expected_cleaned eq REPEAT;

    my $obtained = read_lisp_program(\$input);
    my $dumped   = $obtained->source;
    my $cleaned  = $obtained->clean_tree;

    is_deeply($obtained, $expected,
              "'$input' is parsed correctly");
    is_deeply($dumped, $input,
              "Tree for '$input' is correctly converted back to Lisp code");
    is_deeply($cleaned, $expected_cleaned,
              "Tree for '$input' is correctly cleaned up");
}
